#include <stdio.h>
#include <stdint.h>

#define SIG_MZ      0x5A4D      /* Classic 'MZ' signature. */
#define SIG_ZM      0x4D5A      /* Ancient 'ZM' signature. */

typedef struct {
    uint16_t        signature;      /* MZ or ZM. */
    uint16_t        last_pg;        /* Bytes in last page (or version). */
    uint16_t        n_pages;        /* Number of pages in file. */
    uint16_t        n_relocs;       /* Number of relocations. */
    uint16_t        hdr_paras;      /* Header size in paragraphs. */
    uint16_t        min_alloc;      /* Minimum extra paras needed. */
    uint16_t        max_alloc;      /* Maximum extra paras needed. */
    uint16_t        init_ss;        /* Initial SS value. */
    uint16_t        init_sp;        /* Initial SP value. */
    uint16_t        cksum;          /* Checksum. */
    uint16_t        init_ip;        /* Initial IP value. */
    uint16_t        init_cs;        /* Initial CS value. */
    uint16_t        reloc_ofs;      /* Offset of reloc table. */
    uint16_t        ovl_num;        /* Overlay number. */
} dos_exe_hdr;


#define SIG_RB      0x4252      /* The 'RB' EXEPACK signature. */

int check_dos_exe_file( FILE *f, dos_exe_hdr *hdr )
{
    uint32_t    ep_file_ofs;
    uint16_t    ep_sig;
    uint16_t    insn;
    int         rc;

    /* Calculate the entry point offset in file. */
    ep_file_ofs = (((uint32_t)hdr->init_cs + hdr->hdr_paras) << 4) + hdr->init_ip;
    printf( "Entry point CS:IP      : %X:%X\n", hdr->init_cs, hdr->init_ip );
    printf( "Entry point file offset: %lX\n", ep_file_ofs );

    /* Seek two bytes ahead of the entry point and read the word. */
    fseek( f, ep_file_ofs - sizeof( ep_sig ), SEEK_SET );
    rc = fread( &ep_sig, 1, sizeof( ep_sig ), f );
    if( rc != sizeof( ep_sig ) ) {
        printf( "Read error!\n" );
        return( 10 );
    }

    /* Is the 'RB' signature there? */
    if( ep_sig != SIG_RB ) {
        printf( "Not an EXEPACKed file!\n" );
        return( 20 );
    }
    printf( "EXEPACK signature found\n" );

    /* Now read the first instruction. */
    rc = fread( &insn, 1, sizeof( insn ), f );
    if( rc != sizeof( insn ) ) {
        printf( "Read error!\n" );
        return( 10 );
    }

    if( insn == 0xC08C) {   /* Is it 'MOV AX,ES'? */
        printf( "Destroys initial AX\n" );
    } else if( insn ==0xE88B ) {    /* Is it 'MOV BP,AX'? */
        printf( "Preserves initial AX\n" );
    } else {
        printf( "Don't know this one yet.\n" );
    }

    return( 0 );
}

int main( int argc, char **argv )
{
    int             rc;
    FILE            *f;
    dos_exe_hdr     dos_hdr;

    if( argc != 2) {
        printf( "Usage: %s <exefile>\n", argv[0] );
        return( 1 );
    }

    f = fopen( argv[1], "rb" );
    if( !f ) {
        printf( "Failed to open `%s' for reading!\n", argv[1] );
        return( 2 );
    }

    rc = fread( &dos_hdr, 1, sizeof( dos_hdr ), f );
    if( rc != sizeof( dos_hdr ) ) {
        printf( "Read error!\n" );
        return( 10 );
    }

    if( dos_hdr.signature != SIG_MZ && dos_hdr.signature != SIG_ZM ) {
        printf( "Not a DOS EXE!\n" );
        return( 10 );
    }

    check_dos_exe_file( f, &dos_hdr );

    fclose( f );
    return( 0 );
};
